"""Overlay detected bubbles on the gray stage."""

from itertools import cycle

import cv2 as cv
import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt

from boilercv import DEBUG, PREVIEW, WRITE
from boilercv.captivate.captures import write_image
from boilercv.captivate.previews import view_images
from boilercv.data import VIDEO
from boilercv.data.sets import get_contours_df, get_dataset
from boilercv.examples.previews import _EXAMPLE
from boilercv.images import overlay
from boilercv.images.cv import convert_image, draw_contours
from boilercv.models.params import LOCAL_PATHS
from boilercv.types import ArrInt

_NUM_FRAMES = 1

_PALETTE = [c for c in sns.color_palette("Set1") if not c[0] == c[1] == c[2]]
_PALETTE_CV = [(int(255 * c[2]), int(255 * c[1]), int(255 * c[0])) for c in _PALETTE]


def main():
    frame = _NUM_FRAMES - 1
    gray = (
        get_dataset(_EXAMPLE, _NUM_FRAMES, stage="large_sources")[VIDEO]
        .isel(frame=frame)
        .values
    )
    contours: list[ArrInt] = list(
        get_contours_df(_EXAMPLE)
        .loc[frame, :]
        .groupby("contour")
        .apply(lambda grp: grp.values)
    )
    highlighted = np.zeros_like(convert_image(gray, cv.COLOR_GRAY2BGR))
    for contour, color in zip(contours, cycle(_PALETTE_CV), strict=False):
        highlighted = draw_contours(highlighted, [contour], color=color)
    highlighted = convert_image(highlighted, cv.COLOR_BGR2RGB)
    composed = overlay(gray, highlighted, alpha=0.7)
    if PREVIEW:
        if DEBUG:
            sns.palplot(_PALETTE)
        plt.show()
        view_images(composed)
    if WRITE:
        path = LOCAL_PATHS.media / "examples" / _EXAMPLE / "multicolor"
        path.parent.mkdir(parents=True, exist_ok=True)
        write_image(path, composed)


if __name__ == "__main__":
    main()



"""Handle weird CINEs from 2021-06 to the end of the year.

There is some state being shared between iterations, with the prior frame count
poisoning the next iteration. There's probably a caching issue in the CINE reading
library. Restarting this script after one successful conversion yields another
successful conversion each time.
"""

from loguru import logger

from boilercv.data.video import prepare_dataset
from boilercv.models.params import LOCAL_PATHS
from boilercv.models.paths import get_sorted_paths

FRAMES = {
    "2021-06-21T17-06-43": 3089,
    "2021-08-31T13-59-08": 4411,
    "2021-08-31T14-25-12": 1338,
    "2021-08-31T14-56-17": 3330,
    "2021-08-31T15-13-27": 4176,
    "2021-10-12T12-52-44": 1820,
    "2021-10-12T13-57-43": 4337,
    "2021-10-12T14-37-32": 2368,
    "2021-10-12T17-03-27": 2892,
    "2021-10-13T13-10-59": 3650,
    "2021-12-09T13-04-57": 3455,
}


def main():
    sources = get_sorted_paths(LOCAL_PATHS.data / "weird")
    for source in sources:
        num_frames = FRAMES[source.stem] - 1
        destination = LOCAL_PATHS.large_sources / f"{source.stem}.nc"
        if destination.exists():
            continue
        try:
            prepare_dataset(source, num_frames).to_netcdf(path=destination)
        except Exception:
            logger.exception(source.stem)
            continue


if __name__ == "__main__":
    main()



"""Run examples for the `boilercv.previews` module."""

# ruff: noqa: I001

from pathlib import Path

from boilercv import FRAMERATE_PREV, PREVIEW, WRITE, run_example
from boilercv.captivate.captures import write_video
from boilercv.captivate.previews import view_images
from boilercv.models.params import LOCAL_PATHS
from boilercv.types import DA

from boilercv.previews.gray import main as main1
from boilercv.previews.binarized import main as main2
from boilercv.previews.composite import main as main3
from boilercv.previews.filled import main as main4

results: dict[str, DA] = {}

for func in (main1, main2, main3, main4):
    module_name, result = run_example(func)
    results[module_name] = result
    if WRITE:
        module_path = Path(module_name.replace(".", "/").removeprefix("boilercv/"))
        path = LOCAL_PATHS.media / module_path
        path.parent.mkdir(parents=True, exist_ok=True)
        write_video(path, result, framerate=FRAMERATE_PREV)

if PREVIEW:
    view_images(results, framerate=FRAMERATE_PREV)



"""Examples that operate on large files, which are not in the cloud."""

from collections.abc import Iterator
from contextlib import contextmanager

import xarray as xr

from boilercv.captivate.previews import view_images
from boilercv.data import VIDEO
from boilercv.models.params import LOCAL_PATHS
from boilercv.types import DS

EXAMPLE = LOCAL_PATHS.large_sources / "2022-09-14T13-20-54.nc"


@contextmanager
def example_dataset(
    source: str = "",
    destination: str = "",
    preview: bool = True,
    save: bool = True,
    **kwargs,
) -> Iterator[DS]:  # sourcery skip: move-assign
    """Open the example dataset and write it to a destination file with a suffix.

    Args:
        source: If provided, use a dataset derived from the example with this suffix.
        destination: If provided, add this suffix to the destination dataset file.
        kwargs: Keyword arguments to pass to `xarray.Dataset.to_netcdf`.
        preview: Preview the original and modified datasets.
        save: Whether to save the file.
    """
    _source = (
        LOCAL_PATHS.large_examples / f"{EXAMPLE.stem}_{source}.nc"
        if source
        else EXAMPLE
    )
    _destination = LOCAL_PATHS.large_examples / f"{EXAMPLE.stem}_{destination}.nc"
    with xr.open_dataset(_source) as ds:
        original = ds[VIDEO]
        try:
            yield ds
        finally:
            updated = ds[VIDEO]
            if preview:
                view_images(dict(original=original, updated=updated))
            if save:
                ds.to_netcdf(path=_destination, **kwargs)



"""Save the ROI to a legacy YAML format."""

from warnings import warn

import cv2 as cv

from boilercv.captivate.previews import save_roi
from boilercv.data import apply_to_img_da
from boilercv.data.sets import get_dataset
from boilercv.examples import EXAMPLE_ROI, EXAMPLE_VIDEO_NAME
from boilercv.images import scale_bool
from boilercv.images.cv import find_contours, get_wall
from boilercv.types import DA


def main():
    ds = get_dataset(EXAMPLE_VIDEO_NAME)
    roi = ds["roi"]
    wall: DA = apply_to_img_da(get_wall, scale_bool(roi), name="wall")
    contours = find_contours(scale_bool(wall.values), method=cv.CHAIN_APPROX_SIMPLE)
    if len(contours) > 1:
        warn("More than one contour found when searching for the ROI.", stacklevel=1)
    save_roi(contours[0], EXAMPLE_ROI)



"""Preview a video."""

from pathlib import Path

import cv2 as cv
import numpy as np

from boilercv import PREVIEW
from boilercv.captivate.previews import view_images
from boilercv.examples.cv import capture_images
from boilercv.images.cv import convert_image
from boilercv.types import Vid


def main(preview: bool = PREVIEW) -> Vid:
    images = capture_images(Path(cv.samples.findFile("vtest.avi")))
    video = np.stack([convert_image(image, cv.COLOR_BGR2RGB) for image in images])
    if preview:
        view_images(video)
    return video


if __name__ == "__main__":
    main()



"""Preview results."""

from pathlib import Path

import xarray as xr

from boilercv import DEBUG
from boilercv.colors import RED
from boilercv.data import VIDEO, YX_PX, identity_da
from boilercv.data.sets import slice_frames
from boilercv.images import draw_text, overlay
from boilercv.types import DA, DS

_NUM_FRAMES = 100 if DEBUG else 0


def get_preview(path: Path) -> DS:
    """Get a preview dataset.

    Args:
        path: Path to dataset.
    """
    with xr.open_dataset(path) as ds:
        return xr.Dataset({VIDEO: ds[VIDEO][slice_frames(_NUM_FRAMES)]})


def draw_text_da(da: DA) -> DA:
    """Draw text on images in a data array."""
    frames_dim = str(da.dims[0])
    if da.ndim == 4:
        # We have a color video
        return xr.apply_ufunc(
            draw_text,
            da,
            identity_da(da, frames_dim),
            input_core_dims=([*YX_PX, "channel"], []),
            output_core_dims=([*YX_PX, "channel"],),
            vectorize=True,
        )
    else:
        return xr.apply_ufunc(
            draw_text,
            da,
            identity_da(da, frames_dim),
            input_core_dims=(YX_PX, []),
            output_core_dims=(YX_PX,),
            vectorize=True,
        )


def compose_da(da_image: DA, da_overlay: DA, color: tuple[int, int, int] = RED) -> DA:
    """Draw text on images in a data array.

    Args:
        da_image: Image data array.
        da_overlay: Overlay data array.
        color: Color for the overlay.
    """
    return xr.apply_ufunc(
        overlay,
        da_image,
        da_overlay,
        input_core_dims=(YX_PX, YX_PX),
        output_core_dims=([*YX_PX, "channel"],),
        vectorize=True,
        kwargs=dict(color=color),
    )



"""Types relevant to array manipulation and image processing."""

from typing import Any, Protocol, TypeAlias, TypeVar

import numpy as np
import pandas as pd
import xarray as xr
from numpy import typing as npt

T = TypeVar("T")

DF: TypeAlias = pd.DataFrame
DA: TypeAlias = xr.DataArray
DS: TypeAlias = xr.Dataset


class SupportsMul(Protocol):
    """Protocol for types that support multiplication."""

    def __mul__(self, other: Any) -> Any:
        ...


SupportsMul_T = TypeVar("SupportsMul_T", bound=SupportsMul)

ArrLike: TypeAlias = npt.ArrayLike

NBit: TypeAlias = npt.NBitBase
"""A number with arbitrary precision."""

Arr: TypeAlias = npt.NDArray[np.generic]
"""Generic array type. Consistent with OpenCV's type annotations."""

ArrBool: TypeAlias = npt.NDArray[np.bool_]
"""A boolean array."""

ArrFloat: TypeAlias = npt.NDArray[np.floating[NBit]]
"""An integer array with arbitrary bit depth."""

ArrInt: TypeAlias = npt.NDArray[np.integer[NBit]]
"""An integer array."""

ArrNum: TypeAlias = npt.NDArray[np.number[NBit]]
"""A number array."""

ArrDT: TypeAlias = npt.NDArray[np.datetime64]
"""Datetime array type."""

ArrTD: TypeAlias = npt.NDArray[np.timedelta64]
"""Timedelta array type."""

Img: TypeAlias = ArrInt
"""An integer array representing an image."""

ImgBool: TypeAlias = ArrBool
"""A boolean array representing an image mask."""

ImgLike: TypeAlias = ArrLike
"""An array-like object representable as an image."""

Vid: TypeAlias = Img
"""An integer array representing a video."""

VidBool: TypeAlias = ImgBool
"""A boolean array representing a video mask."""



"""Pack the bits of a NetCDF file."""

from boilercv.data import VIDEO
from boilercv.data.packing import pack
from boilercv.examples.large import example_dataset


def main():
    with example_dataset(source="binarized", destination="packed") as ds:
        ds[VIDEO] = pack(ds[VIDEO])


if __name__ == "__main__":
    main()



"""Bump `pyproject.toml` with changes in `requirements.txt`."""

from pathlib import Path

import toml

PYPROJECT = Path("pyproject.toml")
REQUIREMENTS = Path("requirements.txt")
SOURCE = ".tools" / PYPROJECT
DESTINATION = Path("pyproject.toml")
requirements = REQUIREMENTS.read_text(encoding="utf-8").splitlines()
dependencies = [
    line.rstrip().replace("==", ">=")
    for line in requirements
    if line and not line.startswith("#")
]
source = toml.loads(SOURCE.read_text(encoding="utf-8"))
source["project"]["dependencies"] = dependencies
DESTINATION.write_text(encoding="utf-8", data=toml.dumps(source))



"""Basic examples using OpenCV sample files."""

from collections.abc import Iterator
from os import environ
from pathlib import Path

import cv2 as cv

from boilercv.types import Img


def init():
    """Initialize `boilercv.examples.cv`."""
    check_samples_env_var()


def check_samples_env_var():
    """Check that the OpenCV samples environment variable is set and is a folder."""
    samples_env_var = "OPENCV_SAMPLES_DATA_PATH"
    if (
        not (samples_dir := environ.get(samples_env_var))
        or not Path(samples_dir).is_dir()
    ):
        raise RuntimeError(
            f"{samples_env_var} not set or specified directory does not exist."
        )


def capture_images(path: Path) -> Iterator[Img]:
    """Load images from a video file."""
    video_capture = cv.VideoCapture(str(path))
    while True:
        read_is_successful, image = video_capture.read()
        if not read_is_successful:
            break
        yield image


# * -------------------------------------------------------------------------------- * #

init()



"""Decompress sources to local storage."""


import xarray as xr
from loguru import logger

from boilercv.data import HEADER, VIDEO
from boilercv.data.sets import ALL_NAMES
from boilercv.models.params import LOCAL_PATHS, PARAMS


def main():
    logger.info("start decompress")
    for source_name in ALL_NAMES:
        destination = LOCAL_PATHS.uncompressed_sources / f"{source_name}.nc"
        if destination.exists():
            continue
        source = PARAMS.paths.sources / f"{source_name}.nc"
        with xr.open_dataset(source) as ds:
            xr.Dataset({VIDEO: ds[VIDEO], HEADER: ds[HEADER]}).to_netcdf(destination)
    logger.info("finish decompress")


if __name__ == "__main__":
    main()



"""Preview the binarization stage."""

import xarray as xr

from boilercv import FRAMERATE_PREV, PREVIEW
from boilercv.captivate.previews import view_images
from boilercv.data import VIDEO
from boilercv.images import scale_bool
from boilercv.models.params import PARAMS
from boilercv.previews import draw_text_da
from boilercv.types import DA


def main(preview: bool = PREVIEW) -> DA:
    with xr.open_dataset(PARAMS.paths.binarized_preview) as ds:
        da = draw_text_da(scale_bool(ds[VIDEO]))
    if preview:
        view_images(da, framerate=FRAMERATE_PREV)
    return da


if __name__ == "__main__":
    main()



from pathlib import Path
from typing import TypeVar

import yaml
from pydantic import BaseModel, Extra, MissingError, ValidationError

PARAMS_FILE = Path("params.yaml")

BaseModel_T = TypeVar("BaseModel_T", bound=BaseModel)


class MyBaseModel(BaseModel):
    """Base model and configuration for all models used in this project."""

    class Config:
        """Model configuration. Allow arbitrary types. Enables Numpy types in fields."""

        # Don't specify as class kwargs for easier overriding, and "extra" acted weird.
        extra = Extra.forbid  # To forbid extra fields


# Can't type annotate `model` for some reason
def load_config(path: Path, model: type[BaseModel_T]) -> BaseModel_T:
    """Load a YAML file into a Pydantic model.

    Given a path to a YAML file, automatically unpack its fields into the provided
    Pydantic model.

    Args:
        path: The path to a YAML file.
        model: The model class to pass the contents of the YAML file.

    Returns:
        An instance of the model after validation.

    Raises:
        ValueError: If an improper path is passed.
        ValidationError: If a field is undefined in the configuration file.
    """
    if path.suffix != ".yaml":
        raise ValueError(f"The path '{path}' does not refer to a YAML file.")
    raw_config = yaml.safe_load(path.read_text(encoding="utf-8"))
    if not raw_config:
        raise ValueError("The configuration file is empty.")
    try:
        config = model(**{key: raw_config.get(key) for key in raw_config})
    except ValidationError as exception:
        addendum = "\n  The field may be undefined in the configuration file."
        for error in exception.errors():
            if error["msg"] == MissingError.msg_template:
                error["msg"] += addendum
        raise
    return config


def dump_model(path: Path, model: BaseModel):
    """Dump a Pydantic model to a YAML file.

    Given a path to a YAML file, write a Pydantic model to the file. Optionally add a
    schema directive at the top of the file. Create the file if it doesn't exist.

    Args:
        path: The path to a YAML file. Will create it if it doesn't exist.
        model: An instance of the Pydantic model to dump.
    """
    path = Path(path)
    # ensure one \n and no leading \n, Pydantic sometimes does more
    path.write_text(
        encoding="utf-8",
        data=yaml.safe_dump(model.dict(exclude_none=True), sort_keys=False),
    )


def write_schema(path: Path, model: type[BaseModel]):
    """Write a Pydantic model schema to a JSON file.

    Given a path to a JSON file, write a Pydantic model schema to the file. Create the
    file if it doesn't exist.

    Args:
        path: The path to a JSON file. Will create it if it doesn't exist.
        model: The Pydantic model class to get the schema from.
    """
    if path.suffix != ".json":
        raise ValueError(f"The path '{path}' does not refer to a JSON file.")
    path.write_text(model.schema_json(indent=2) + "\n", encoding="utf-8")



"""Get bubble contours."""

import cv2 as cv

from boilercv.captivate.previews import view_images
from boilercv.data import IDX, VIDEO
from boilercv.data.sets import get_dataset
from boilercv.examples import EXAMPLE_CONTOURS, EXAMPLE_NUM_FRAMES, EXAMPLE_VIDEO_NAME
from boilercv.images import scale_bool
from boilercv.images.cv import draw_contours
from boilercv.stages.contours import get_all_contours
from boilercv.types import ArrInt, Img


def main():
    ds = get_dataset(EXAMPLE_VIDEO_NAME, EXAMPLE_NUM_FRAMES)
    video = ds[VIDEO]
    df = get_all_contours(
        cv.bitwise_not(scale_bool(video.values)), method=cv.CHAIN_APPROX_SIMPLE
    )
    df.to_hdf(EXAMPLE_CONTOURS, "contours", complib="zlib", complevel=9)
    result: list[Img] = []
    for frame_num, frame in enumerate(video):
        contours: list[ArrInt] = list(  # type: ignore
            df.loc[IDX[frame_num], :]  # type: ignore
            .groupby("contour")
            .apply(lambda grp: grp.values)  # type: ignore
        )
        result.append(draw_contours(scale_bool(frame.values), contours))
    view_images(result)


if __name__ == "__main__":
    main()



"""Overlay detected bubbles on the binarized stage."""

from boilercv import FRAMERATE_CONT, PREVIEW, WRITE
from boilercv.captivate.captures import write_video
from boilercv.captivate.previews import view_images
from boilercv.data import VIDEO
from boilercv.data.sets import get_dataset
from boilercv.examples.previews import _EXAMPLE, _NUM_FRAMES
from boilercv.images import scale_bool
from boilercv.models.params import LOCAL_PATHS
from boilercv.previews import compose_da


def main():
    source = get_dataset(_EXAMPLE, _NUM_FRAMES, stage="sources")[VIDEO]
    bubbles = get_dataset(_EXAMPLE, _NUM_FRAMES, stage="filled")[VIDEO]
    highlighted_bubbles = compose_da(source, scale_bool(bubbles)).transpose(
        "frame", "ypx", "xpx", "channel"
    )
    if PREVIEW:
        view_images(bubbles.isel(frame=0))
        view_images(highlighted_bubbles, framerate=FRAMERATE_CONT)
    if WRITE:
        write_video(LOCAL_PATHS.media / "binarized", source)
        write_video(LOCAL_PATHS.media / "bubbles", bubbles, preview_frame=True)
        write_video(LOCAL_PATHS.media / "binarized_highlighted", highlighted_bubbles)


if __name__ == "__main__":
    main()



"""Detect the boiling surface."""

from warnings import warn

import cv2 as cv
import numpy as np
import xarray as xr
from scipy.ndimage import (
    center_of_mass,
    generate_binary_structure,
    label,
    labeled_comprehension,
)

from boilercv import DEBUG
from boilercv.captivate.previews import get_calling_scope_name, save_roi, view_images
from boilercv.data import VIDEO, YX_PX, apply_to_img_da
from boilercv.data.frames import df_points
from boilercv.data.sets import get_dataset
from boilercv.examples import EXAMPLE_NUM_FRAMES, EXAMPLE_ROI, EXAMPLE_VIDEO_NAME
from boilercv.images import scale_bool
from boilercv.images.cv import find_contours, get_wall
from boilercv.types import DA, ArrInt, Img


def main():
    ds = get_dataset(EXAMPLE_VIDEO_NAME, EXAMPLE_NUM_FRAMES)
    video = ds[VIDEO]
    roi = ds["roi"]
    wall: DA = apply_to_img_da(get_wall, scale_bool(roi), name="wall")
    boiling_surface, boiling_surface_coords = xr.apply_ufunc(
        find_boiling_surface,
        scale_bool(wall),
        input_core_dims=[YX_PX],
        output_core_dims=[YX_PX, ["test"]],
    )
    contours = find_contours(scale_bool(wall.values), method=cv.CHAIN_APPROX_SIMPLE)
    if len(contours) > 1:
        warn("More than one contour found when searching for the ROI.", stacklevel=1)
    save_roi(contours[0], EXAMPLE_ROI)
    view_images(dict(boiling_surface=boiling_surface, roi=roi))


def find_boiling_surface(img: Img) -> tuple[Img, ArrInt]:
    """Find the boiling surface."""

    # Parameters for finding prominent horizontal lines
    wide_rectangular_ksize = [9, 3]  # Enhance horizontal lines
    threshold = 0.6

    # The range to consider a connected component to be the boiling surface
    (height, width) = img.shape
    min_size_px = 8
    ypx_min = height / 2
    xpx_mid = width / 2
    xpx_max_dist = width / 8
    xpx_min = xpx_mid - xpx_max_dist
    xpx_max = xpx_mid + xpx_max_dist

    # Find prominent horizontal lines
    corners = cv.cornerHarris(src=img, blockSize=2, ksize=3, k=0.04)
    lines = -1 * corners  # Linear features are strongly negative in Harris
    blurred = cv.blur(lines, ksize=wide_rectangular_ksize)
    scaled = (blurred - blurred.min()) / (blurred.max() - blurred.min())
    binarized = scaled > threshold

    # Find connected components and their centers
    # ? cv.connectedComponents and cv.moments could also be used
    # ? Consider re-implementing if slow
    find_diag_conns = generate_binary_structure(rank=2, connectivity=2)
    labeled_img, num_objects = label(input=binarized, structure=find_diag_conns)
    labels = range(1, num_objects + 1)  # Exclude 0-labeled background
    sizes_px = labeled_comprehension(
        input=binarized,
        labels=labeled_img,
        index=labels,
        func=np.count_nonzero,
        out_dtype=int,
        default=0,
    )
    objs = (
        df_points(
            np.array(center_of_mass(input=binarized, labels=labeled_img, index=labels))
            .round()
            .astype(int)  # Not img dtype. df contains coords, not pixel values
        )
        .assign(**dict(label=labels, size_px=sizes_px))
        .rename(axis="columns", mapper=dict(ypx="ypx_center", xpx="xpx_center"))
        .set_index("label", drop=True)
    )

    # Consider the boiling surface to be the largest object in range
    objs_in_range = objs[
        (objs.size > min_size_px)
        & (objs.ypx_center > ypx_min)
        & (objs.xpx_center > xpx_min)
        & (objs.xpx_center < xpx_max)
    ]
    label_largest_in_range = objs_in_range.size_px.idxmax()
    boiling_surface = labeled_img == label_largest_in_range

    # Get a line segment corresponding to the boiling surface
    points = df_points(np.array(boiling_surface.nonzero()).T)
    xpx_left = points.xpx.min()
    xpx_right = points.xpx.max()
    ypx_horizontal = np.round(points.ypx.mean()).astype(int)
    boiling_surface_coords = np.array(
        [ypx_horizontal, xpx_left, ypx_horizontal, xpx_right]
    ).astype(int)

    if DEBUG:
        view_images(
            name=get_calling_scope_name(),
            images=dict(
                img=img,
                corners=corners,
                lines=lines,
                blurred=blurred,
                scaled=scaled,
                binarized=binarized,
                labeled_img=labeled_img,
                boiling_surface=boiling_surface,
            ),
        )
    return boiling_surface, boiling_surface_coords


if __name__ == "__main__":
    main()



"""Tools for datasets."""

from collections.abc import Callable, Sequence
from itertools import chain
from typing import Any

import numpy as np
import pandas as pd
import xarray as xr
from pytz import timezone

from boilercv.data.models import Dimension, get_dims
from boilercv.types import DA, DS, ArrLike

VIDEO = "video"
ROI = "roi"
HEADER = "header"
TIMEZONE = timezone("US/Pacific")

FRAME = "frame"
TIME = "time"
UTC_TIME = "utc"

Y = "y"
X = "x"
YX = ["y", "x"]

PX = "px"
YPX = f"{Y}{PX}"
XPX = f"{X}{PX}"
YX_PX = [YPX, XPX]

DIMS = [FRAME, YPX, XPX]

PACKED_DIM_INDEX = 2
PACKED = "packed"
XPX_PACKED = "xpx_packed"
PACKED_DIMS = [FRAME, YPX, XPX_PACKED]

VIDEO_NAME = "video_name"

LENGTH = "um"
SAMPLE_DIAMETER_UM = 9_525_000
ROI = "roi"
OTHER_ROI = "roi_other"

IDX = pd.IndexSlice
"""Helper for slicing multi-index dataframes."""


def identity_da(da: DA, dim: str) -> DA:
    """Construct a data array that maps a dimension's coordinates to itself.

    Useful to apply `xr.apply_ufunc` along coordinate values.

    Args:
        da: Data array.
        dim: The dimension to extract.
    """
    return xr.DataArray(
        dims=(dim),
        coords={dim: da[dim].values},
        data=da[dim],
    )


def apply_to_img_da(
    func: Callable[..., Any],
    *args: Any,
    returns: int | None = 1,
    vectorize: bool = False,
    name: Sequence[str] | str = "",
    kwargs: dict[str, Any] | None = None,
) -> Any:
    """Apply functions that transform images to transform data arrays instead."""
    core_dims = [YX_PX]
    common_kwargs = dict(
        input_core_dims=core_dims * len(args),
        vectorize=vectorize,
        keep_attrs=True,
        kwargs=kwargs,
    )
    if returns and returns > 0:
        result = xr.apply_ufunc(
            func,
            *args,
            **common_kwargs,
            output_core_dims=core_dims * returns,
        )
        if name and returns == 1:
            result = result.rename(name)
        if name and returns > 1:
            result = tuple(
                part.rename(name_) for name_, part in zip(name, result, strict=True)
            )
        return result
    if not returns or returns == 0:
        xr.apply_ufunc(
            func,
            *args,
            **common_kwargs,
        )


def assign_ds(
    name: str,
    data: ArrLike,
    dims: Sequence[Dimension],
    secondary_dims: Sequence[Dimension] = (),
    fixed_dims: Sequence[Dimension] = (),
    fixed_secondary_dims: Sequence[Dimension] = (),
    ds: DS | None = None,
    long_name: str = "",
    units: str = "",
) -> DS:
    """Build a data array and assign it to a dataset."""
    if not ds:
        ds = xr.Dataset()
    da = build_da(
        name,
        data,
        dims,
        secondary_dims,
        fixed_dims,
        fixed_secondary_dims,
        long_name,
        units,
    )
    ds[name] = da
    return ds


def build_da(
    name: str,
    data: ArrLike,
    dims: Sequence[Dimension],
    secondary_dims: Sequence[Dimension] = (),
    fixed_dims: Sequence[Dimension] = (),
    fixed_secondary_dims: Sequence[Dimension] = (),
    long_name: str = "",
    units: str = "",
):
    """Build a data array."""
    all_primary_dims = chain(fixed_dims, dims)
    all_fixed_dims = chain(fixed_dims, fixed_secondary_dims)
    all_variable_dims = chain(dims, secondary_dims)
    if not long_name:
        long_name = name.capitalize() if len(name) > 1 else name
    if isinstance(data, Sequence) and not len(data):
        shape: list[int] = []
        for dim in all_primary_dims:
            if dim.coords is None:
                shape.append(1)
            else:
                shape.append(len(dim.coords))
        data = np.full(shape, None)
    attrs = {"long_name": long_name}
    if units:
        attrs["units"] = units
    da = xr.DataArray(
        name=name,
        dims=get_dims(*all_primary_dims),
        data=data,
        attrs=attrs,
    )
    for dim in chain(all_fixed_dims, all_variable_dims):
        da = dim.assign_to(da)
    return da



"""Perform all of the steps."""

from boilercv import DEBUG
from boilercv.captivate.previews import view_images
from boilercv.data import FRAME, VIDEO, apply_to_img_da
from boilercv.data.packing import pack
from boilercv.examples.large import example_dataset
from boilercv.images import scale_bool
from boilercv.images.cv import apply_mask, binarize, flood, get_roi
from boilercv.types import DA


def main():
    with example_dataset(
        destination="combined", preview=DEBUG, encoding={VIDEO: {"zlib": True}}
    ) as ds:
        video = ds[VIDEO]
        maximum = video.max(FRAME)
        flooded: DA = apply_to_img_da(flood, maximum)
        roi = apply_to_img_da(get_roi, scale_bool(flooded))
        masked: DA = apply_to_img_da(apply_mask, video, scale_bool(roi), vectorize=True)
        binarized: DA = apply_to_img_da(binarize, masked, vectorize=True)
        ds[VIDEO] = pack(binarized)
        view_images(
            dict(
                video=video.isel(frame=0),
                maximum=maximum,
                flooded=flooded,
                roi=roi,
                masked=masked.isel(frame=0),
                binarized=binarized.isel(frame=0),
            )
        )


if __name__ == "__main__":
    main()



"""Flatten the data directory structure.

The directory structure looks like:

    data
    └───YYYY-MM-DD
        ├───data
        ├───notes
        └───video
"""

from itertools import chain

from boilercv.models.params import LOCAL_PATHS


def main():
    source = LOCAL_PATHS.hierarchical_data
    rename_notes(source)
    rename_cines(source)
    rename_sheets(source)


def rename_notes(source):
    notes_dest = LOCAL_PATHS.notes
    notes_dirs = {
        trial.stem: trial / "notes"
        for trial in sorted(source.iterdir())
        if trial.is_dir()
    }
    for trial, note_dir in notes_dirs.items():
        if not note_dir.exists():
            continue
        note_dir.rename(notes_dest / trial)


def rename_cines(source):
    destination = LOCAL_PATHS.cines
    trials = [trial / "video" for trial in source.iterdir() if trial.is_dir()]
    videos = chain.from_iterable(trial.glob("*.cine") for trial in trials)
    for video in videos:
        video.rename(destination / video.name.removeprefix("results_"))


def rename_sheets(source):
    sheets_dest = LOCAL_PATHS.sheets
    data = [trial / "data" for trial in sorted(source.iterdir()) if trial.is_dir()]
    sheets = chain.from_iterable(trial.glob("*.csv") for trial in data)
    for sheet in sheets:
        sheet.rename(sheets_dest / sheet.name.removeprefix("results_"))


if __name__ == "__main__":
    main()



from collections.abc import Sequence

import cv2 as cv
import numpy as np

from boilercv.colors import WHITE, WHITE3
from boilercv.images import unpad
from boilercv.types import ArrFloat, ArrInt, Img, ImgBool


def convert_image(img: Img, code: int | None = None) -> Img:
    """Convert image format, handling inconsistent type annotations."""
    return cv.cvtColor(img, code)  # type: ignore


def apply_mask(img: Img, mask: Img) -> Img:
    """Mask an image, keeping parts where the mask is bright."""
    return cv.add(img, cv.bitwise_not(mask))


def pad(img: Img, pad_width: int, value: int) -> Img:
    """Pad an image. Faster than np.pad()."""
    return cv.copyMakeBorder(
        img,
        top=pad_width,
        bottom=pad_width,
        left=pad_width,
        right=pad_width,
        borderType=cv.BORDER_CONSTANT,
        value=value,
    )


def binarize(img: Img, block_size: int = 11, thresh_dist_from_mean: int = 2) -> ImgBool:
    """Binarize an image with an adaptive threshold."""
    block_size += 1 if block_size % 2 == 0 else 0
    return cv.adaptiveThreshold(
        src=img,
        maxValue=np.iinfo(img.dtype).max,
        adaptiveMethod=cv.ADAPTIVE_THRESH_MEAN_C,
        thresholdType=cv.THRESH_BINARY,
        blockSize=block_size,
        C=thresh_dist_from_mean,
    ).astype(bool)


def flood(img: Img) -> ImgBool:
    """Flood the image, returning the resulting flood as a bright mask."""
    seed_point = np.array(img.shape) // 2
    max_value = np.iinfo(img.dtype).max
    # OpenCV needs a masked array with a one-pixel pad
    pad_width = 1
    mask = pad(np.zeros_like(img), pad_width, value=1)
    _retval, _image, mask, _rect = cv.floodFill(
        image=img,
        mask=mask,
        seedPoint=seed_point,
        newVal=None,  # Ignored in mask only mode
        flags=cv.FLOODFILL_MASK_ONLY,
    )
    # Return the mask in original dimensions
    return unpad(mask, pad_width).astype(bool)


def get_roi(wall: Img) -> Img:
    """Erode the wall to get the ROI."""
    kernel_size = 9
    close_kernel_size = 4
    kernel = cv.getStructuringElement(cv.MORPH_ELLIPSE, [kernel_size] * 2)
    close_kernel = cv.getStructuringElement(cv.MORPH_ELLIPSE, [close_kernel_size] * 2)
    # Explicitly pad out the image since cv.morphologyEx boundary handling is weird
    pad_width = max(close_kernel_size, kernel_size)
    padded = pad(wall, pad_width, value=0)
    wall = cv.morphologyEx(src=padded, op=cv.MORPH_CLOSE, kernel=close_kernel)
    roi = cv.morphologyEx(src=wall, op=cv.MORPH_ERODE, kernel=kernel)
    return unpad(roi, pad_width).astype(bool)


def get_wall(roi: Img) -> Img:
    """Dilate the ROI to get the wall."""
    # Explicitly pad out the image since cv.morphologyEx boundary handling is weird
    kernel_size = 9
    kernel = cv.getStructuringElement(cv.MORPH_ELLIPSE, [kernel_size] * 2)
    pad_width = kernel_size
    padded = pad(roi, pad_width, value=0)
    wall = cv.morphologyEx(src=padded, op=cv.MORPH_DILATE, kernel=kernel)
    return unpad(wall, pad_width).astype(bool)


def build_mask_from_polygons(img: Img, contours: Sequence[ArrInt]) -> Img:
    """Build a mask from the intersection of a sequence of polygonal contours."""
    # OpenCV expects contours as shape (N, 1, 2) instead of (N, 2)
    contours = [np.fliplr(contour).reshape(-1, 1, 2) for contour in contours]
    blank = np.zeros_like(img)
    return cv.fillPoly(
        img=blank,
        pts=contours,  # Expects a list of coordinates, we have just one
        color=WHITE3,
    )


def find_contours(img: Img, method: int = cv.CHAIN_APPROX_NONE) -> list[ArrInt]:
    """Find external contours of bright objects in an image."""
    contours, _hierarchy = cv.findContours(
        image=img,
        mode=cv.RETR_EXTERNAL,  # No hierarchy needed because we keep external contours
        method=method,
    )
    # Despite images having dims (y, x) and shape (h, w), OpenCV returns contours with
    # dims (point, 1, pair), where dim "pair" has coords (x, y).
    contours = [np.fliplr(contour.reshape(-1, 2)) for contour in contours]
    return contours


def draw_contours(
    img: Img,
    contours: Sequence[ArrInt],
    contour_index: int = -1,
    thickness: int = -1,
    color: int | tuple[int, ...] = WHITE,
) -> Img:
    """Draw contours on an image."""
    # OpenCV expects contours as shape (N, 1, 2) instead of (N, 2)
    contours = [np.fliplr(contour).reshape(-1, 1, 2) for contour in contours]
    return cv.drawContours(
        image=img,
        contours=contours,
        contourIdx=contour_index,
        color=color,
        thickness=thickness,
    )


def find_line_segments(img: Img) -> tuple[ArrFloat, cv.LineSegmentDetector]:
    """Find line segments in an image."""
    lsd = cv.createLineSegmentDetector()
    lines, *_ = lsd.detect(img)
    # OpenCV returns line segments as shape (N, 1, 4) instead of (N, 4)
    lines = lines.reshape(-1, 4)
    return lines, lsd



"""Fill bubble contours."""

import cv2 as cv
import numpy as np
import xarray as xr
from loguru import logger

from boilercv.data import IDX, ROI, VIDEO
from boilercv.data.packing import pack
from boilercv.data.sets import get_all_datasets, get_contours_df
from boilercv.images.cv import draw_contours
from boilercv.models.params import PARAMS
from boilercv.types import ArrInt


def main():
    for ds, name in get_all_datasets():
        destination = PARAMS.paths.filled / f"{name}.nc"
        if destination.exists():
            continue

        df = get_contours_df(name)
        ds = xr.zeros_like(ds, dtype=np.uint8)
        video = ds[VIDEO]
        if not df.empty:
            for frame_num, frame in enumerate(video):
                try:
                    video[frame_num, :, :] = loop(df, frame_num, frame)
                except Exception:
                    logger.exception("Failed to fill contours")
                    continue
        ds[VIDEO] = pack(video)
        ds = ds.drop_vars(ROI)
        ds.to_netcdf(path=destination, encoding={VIDEO: {"zlib": True}})


def loop(df, frame_num, frame):
    contours: list[ArrInt] = list(  # type: ignore
        df.loc[IDX[frame_num], :]  # type: ignore
        .groupby("contour")
        .apply(lambda grp: grp.values)  # type: ignore
    )
    return draw_contours(frame.values, contours, thickness=cv.FILLED)


if __name__ == "__main__":
    logger.info("start fill contours")
    main()
    logger.info("finish fill contours")



"""Binarize all videos and export their ROIs."""

import xarray as xr
from loguru import logger

from boilercv.data import FRAME, ROI, VIDEO, apply_to_img_da
from boilercv.data.packing import pack
from boilercv.images import scale_bool
from boilercv.images.cv import apply_mask, binarize, flood, get_roi
from boilercv.models.params import LOCAL_PATHS, PARAMS
from boilercv.models.paths import get_sorted_paths
from boilercv.types import DA


def main():
    logger.info("start binarize")
    for source in get_sorted_paths(LOCAL_PATHS.large_sources):
        destination = PARAMS.paths.sources / f"{source.stem}.nc"
        if destination.exists():
            continue
        with xr.open_dataset(source) as ds:
            video = ds[VIDEO]
            maximum = video.max(FRAME)
            flooded: DA = apply_to_img_da(flood, maximum)
            roi: DA = apply_to_img_da(get_roi, scale_bool(flooded))
            masked: DA = apply_to_img_da(
                apply_mask, video, scale_bool(roi), vectorize=True
            )
            binarized: DA = apply_to_img_da(binarize, masked, vectorize=True)
            ds[VIDEO] = pack(binarized)
            ds.to_netcdf(
                path=PARAMS.paths.sources / source.name,
                encoding={VIDEO: {"zlib": True}},
            )
            ds[ROI] = roi
            ds = ds.drop_vars(VIDEO)
            ds.to_netcdf(path=PARAMS.paths.rois / source.name)
    logger.info("finish binarize")


if __name__ == "__main__":
    main()



"""Preview detected regions on an image from the dataset."""

import xarray as xr

from boilercv import PREVIEW, WRITE
from boilercv.captivate.captures import write_image
from boilercv.captivate.previews import view_images
from boilercv.colors import BLUE, GREEN, RED
from boilercv.data import ROI, VIDEO
from boilercv.data.sets import get_dataset
from boilercv.examples.detect_surface import find_boiling_surface
from boilercv.examples.previews import _EXAMPLE
from boilercv.images import overlay, scale_bool
from boilercv.models.params import LOCAL_PATHS, PARAMS


def main():
    ds = get_dataset(_EXAMPLE)
    gray = (
        xr.open_dataset(PARAMS.paths.gray_preview)[VIDEO]
        .sel(video_name=_EXAMPLE)
        .values
    )
    roi = ds[ROI].values
    highlighted_roi = overlay(gray, scale_bool(roi), color=BLUE, alpha=0.2)

    surface, _ = find_boiling_surface(scale_bool(roi))
    highlighted_surface = overlay(
        highlighted_roi, scale_bool(surface), color=RED, alpha=1
    )

    filled = (
        xr.open_dataset(PARAMS.paths.filled_preview)[VIDEO]
        .sel(video_name=_EXAMPLE)
        .values
    )
    highlighted_bubbles = overlay(
        highlighted_surface, scale_bool(filled), color=GREEN, alpha=0.4
    )

    if PREVIEW:
        view_images(highlighted_bubbles)
    if WRITE:
        write_image(LOCAL_PATHS.media / "roi", highlighted_roi)
        write_image(LOCAL_PATHS.media / "composite", highlighted_bubbles)


if __name__ == "__main__":
    main()



"""Old approach to detecting the boiling surface."""

import numpy as np
import pandas as pd

from boilercv import DEBUG
from boilercv.captivate.previews import view_images
from boilercv.data import YX_PX, apply_to_img_da
from boilercv.data.frames import df_points, frame_lines
from boilercv.examples.detect_surface import find_boiling_surface
from boilercv.images.cv import find_contours, find_line_segments
from boilercv.types import DA, ArrInt


def _find_boiling_surface(image: DA, preview: bool = DEBUG) -> ArrInt:
    """Find the boiling surface."""
    # Find corners and set index to contour points for later concatenation
    (height, width) = image.shape
    corns_ = apply_to_img_da(find_boiling_surface, image)
    corns = df_points(np.nonzero(corns_.values))
    corns = (
        corns.set_index(pd.MultiIndex.from_frame(corns))
        .drop(axis="columns", labels=YX_PX)
        .assign(**dict(corner=True))
    )
    # Find the contour angles
    contours = find_contours(image.values)
    roi_poly_ = contours.pop()
    roi_poly = df_points(roi_poly_)
    distances = roi_poly.diff()
    angles = (
        pd.Series(np.degrees(np.arctan2(distances.ypx, distances.xpx)))
        .rolling(window=2)
        .mean()
        .bfill()
    ).rename("deg")
    # Concatenate contour points with their angles
    roi_poly = (
        pd.concat(
            axis="columns",
            keys=["dim", "angle"],
            objs=[roi_poly, angles],
        )
        .set_index(pd.MultiIndex.from_frame(roi_poly))
        .drop(axis="columns", labels="dim")
    )
    # Set the index to the contour points and concatenate with corners
    # Compute distance from midplanes for candidate corner selection
    roi_poly = roi_poly.set_axis(axis="columns", labels=roi_poly.columns.droplevel(1))
    corns = (
        pd.concat(axis="columns", objs=[roi_poly, corns])
        .dropna()
        .drop(axis="columns", labels="corner")
        .sort_index()
        .reset_index()
        .assign(
            **dict(
                xpx_mid_abs=lambda df: (df.xpx - width / 2).abs(),
                ypx_mid=lambda df: df.ypx - height / 2,
            )
        )
        .sort_values(axis="index", by="xpx_mid_abs")
    )
    # Consider the boiling surface to be described by the two corners with a low enough
    # angle, below the horizontal midplane, and closest to the vertical midplane
    max_angle = 30
    candidates = (
        corns[
            (corns.angle > -max_angle) & (corns.angle < max_angle) & (corns.ypx_mid > 0)
        ]
        .reset_index(drop=True)
        .loc[0:1, ["ypx", "xpx"]]
    )
    if preview:
        view_images(dict(corn=corns_.values))
    # TODO: Save to disk, draw on an image for previewing
    return candidates.values.flatten()[:2]


def _find_boiling_surface2(image):
    """Get line segments in an image."""
    lines_, lsd = find_line_segments(image)
    lined = lsd.drawSegments(image, lines_)
    lines = frame_lines(lines_)
    midpoints = df_points([lines.ypx.T.mean(), lines.xpx.T.mean()])
    distances = df_points([lines.ypx[1] - lines.ypx[0], lines.xpx[1] - lines.xpx[0]])
    lengths = distances.T.pow(2).sum().pow(1 / 2).rename("px")
    angles = pd.Series(np.arctan2(distances.ypx, distances.xpx)).rename("deg")
    lines = (
        pd.concat(
            axis="columns",
            keys=["line", "midpoint", "length", "angle"],
            objs=[
                lines.set_axis(axis="columns", labels=["ypx0", "xpx0", "ypx1", "xpx1"]),
                midpoints,
                lengths,
                angles,
            ],
        )
        .rename_axis(axis="index", mapper="line")
        .rename_axis(axis="columns", mapper=["metric", "dim"])
    )
    return lined, lines



"""Overlay detected bubbles on the image."""

from boilercv import FRAMERATE_PREV, PREVIEW, WRITE
from boilercv.captivate.captures import write_video
from boilercv.captivate.previews import view_images
from boilercv.data import VIDEO
from boilercv.images import scale_bool
from boilercv.models.params import LOCAL_PATHS, PARAMS
from boilercv.previews import compose_da, draw_text_da, get_preview
from boilercv.types import DA


def main(preview: bool = PREVIEW) -> DA:
    gray = get_preview(PARAMS.paths.gray_preview)
    filled = get_preview(PARAMS.paths.filled_preview)
    composed = draw_text_da(
        compose_da(gray[VIDEO], scale_bool(filled[VIDEO])).transpose(
            "video_name", "ypx", "xpx", "channel"
        )
    )
    if preview:
        view_images(composed, framerate=FRAMERATE_PREV)
    if WRITE:
        write_video(
            LOCAL_PATHS.media / "examples/composite", composed, framerate=FRAMERATE_PREV
        )
    return composed


if __name__ == "__main__":
    main()



"""Run examples."""

from typing import Any

from boilercv import PREVIEW, run_example
from boilercv.captivate.previews import view_images
from boilercv.examples.cv.basic_test import main as main2
from boilercv.examples.cv.starry import main as main1

results: dict[str, Any] = {}

for func in (main1, main2):
    module_name, result = run_example(func)
    results[module_name] = result

if PREVIEW:
    view_images(results)



"""Show an OpenCV sample image."""

import cv2 as cv

from boilercv import PREVIEW
from boilercv.captivate.previews import view_images
from boilercv.types import ImgLike


def main(preview=PREVIEW) -> ImgLike:
    img = cv.cvtColor(
        cv.imread(cv.samples.findFile("starry_night.jpg")), cv.COLOR_BGR2RGB
    )
    if preview:
        view_images(img)
    return img


if __name__ == "__main__":
    main()



"""Generate schema for the project."""

import re

from boilercv.models import write_schema
from boilercv.models.params import PARAMS, Params


def main():
    write_schema(
        PARAMS.paths.project_schema / f"{to_snake_case(Params.__name__)}_schema.json",
        Params,
    )


# https://github.com/samuelcolvin/pydantic/blob/4f4e22ef47ab04b289976bb4ba4904e3c701e72d/pydantic/utils.py#L127-L131
def to_snake_case(v: str) -> str:
    v = re.sub(r"([A-Z]+)([A-Z][a-z])", r"\1_\2", v)
    v = re.sub(r"([a-z\d])([A-Z])", r"\1_\2", v)
    return v.replace("-", "_").lower()


if __name__ == "__main__":
    main()



"""Video dataset model."""

from dataclasses import asdict
from pathlib import Path

import xarray as xr
from boilercine import get_cine_attributes, get_cine_images
from scipy.spatial.distance import euclidean

from boilercv.captivate.previews import load_roi
from boilercv.data import (
    FRAME,
    HEADER,
    LENGTH,
    SAMPLE_DIAMETER_UM,
    TIME,
    TIMEZONE,
    UTC_TIME,
    VIDEO,
    XPX,
    YPX,
    YX,
    assign_ds,
)
from boilercv.data.models import Dimension
from boilercv.models.params import PARAMS
from boilercv.types import DA, DS


def prepare_dataset(
    cine_source: Path, num_frames: int | None = None, start_frame: int = 0
) -> DS:
    """Prepare a dataset from a CINE."""

    # Header
    header, utc_arr = get_cine_attributes(
        cine_source, TIMEZONE, num_frames, start_frame
    )
    header_da = xr.DataArray(name=HEADER, attrs=asdict(header))

    # Dimensions
    frame_dim = Dimension(
        dim=FRAME,
        long_name="Frame number",
    )
    time = Dimension(
        parent_dim=frame_dim.dim,
        dim=TIME,
        long_name="Time elapsed",
        units="s",
        original_units="ns",
        original_coords=(utc_arr - utc_arr[0]).astype(float),
        scale=1e-9,
    )
    utc = Dimension(
        parent_dim=frame_dim.dim,
        dim=UTC_TIME,
        long_name="UTC time",
        coords=utc_arr,
    )

    # Dataset
    ds = assign_ds(
        name=VIDEO,
        long_name="High-speed video data",
        units="Pixel intensity",
        fixed_dims=(frame_dim,),
        dims=(
            Dimension(
                dim=YPX,
                long_name="Height",
                units="px",
            ),
            Dimension(
                dim=XPX,
                long_name="Width",
                units="px",
            ),
        ),
        fixed_secondary_dims=(time, utc),
        data=list(get_cine_images(cine_source, num_frames, start_frame)),
    )
    ds[header_da.name] = header_da
    return ds


# * -------------------------------------------------------------------------------- * #
# * SECONDARY LENGTH DIMENSIONS


def assign_length_dims(dataset: DS) -> DS:
    """Assign length scales to "x" and "y" coordinates."""
    images = dataset[VIDEO]
    parent_dim_units = "px"
    roi = load_roi(images.data, PARAMS.paths.examples / "roi_line.yaml", "line")
    pixels = euclidean(*iter(roi))
    um_per_px = SAMPLE_DIAMETER_UM / pixels
    y = get_length_dims(parent_dim_units, YX[0], "Height", um_per_px, images)
    x = get_length_dims(parent_dim_units, YX[1], "Width", um_per_px, images)
    images = y.assign_to(images)
    images = x.assign_to(images)
    return dataset


def get_length_dims(
    parent_dim_units: str, dim: str, long_name: str, scale: float, images: DA
) -> Dimension:
    """Get a length scale."""
    parent_dim = f"{dim}{parent_dim_units}"
    return Dimension(
        parent_dim=f"{dim}{parent_dim_units}",
        dim=dim,
        long_name=long_name,
        units=LENGTH,
        original_units=parent_dim_units,
        original_coords=images[parent_dim].values,
        scale=scale,
    )



"""Computer vision routines suitable for nucleate pool boiling bubble analysis."""

from collections.abc import Callable
from os import environ
from typing import Any

import pandas as pd
from cv2 import version
from loguru import logger

_debug = environ.get("BOILERCV_DEBUG")
_preview = environ.get("BOILERCV_PREVIEW")
_write = environ.get("BOILERCV_WRITE")
DEBUG = str(_debug).casefold() == "true" if _debug else False
"""Whether to run in debug mode. Log to `boilercv.log`."""
PREVIEW = str(_preview).casefold() == "true" if _preview else False
"""Whether to run interactive previews."""
WRITE = str(_write).casefold() == "true" if _write else False
"""Whether to write to the local media folder."""

FFMPEG_LOG_LEVEL = "warn" if DEBUG else "error"
"""Log level for FFMPEG."""
FRAMERATE_CONT = 60
"""Framerate for continuous video output."""
FRAMERATE_PREV = 3
"""Framerate for previews from multiple results, as in a slideshow."""


def init():
    """Initialize `boilercv`."""
    if DEBUG:
        logger.add(sink="boilercv.log")
    check_contrib()
    pd.set_option("mode.copy_on_write", True)
    pd.set_option("mode.chained_assignment", "raise")
    pd.set_option("mode.string_storage", "pyarrow")


_CONTRIB_MSG = """\
OpenCV is not installed with extras. A dependent package may have pinned `opencv-pyhon`
and clobbered your installed version.
"""


def check_contrib():
    """Ensure the installed version of OpenCV has extras.

    Dependencies can specify a different version of OpenCV than the one required in this
    project, unintentionally clobbering the installed version of OpenCV. Detect whether
    a non-`contrib` version is installed by a dependency.
    """
    if not version.contrib:
        raise ImportError(_CONTRIB_MSG)


def run_example(func: Callable[..., Any], preview: bool = False) -> tuple[str, Any]:
    """Run an example file, logging the module name containing the function.

    Args:
        func: The example function to run.
        preview: Preview results from the function. Default: False.
    """
    module_name = func.__module__
    logger.info(f'Running example "{module_name}"')
    result = func(preview=preview)
    return module_name, result


init()



"""Get bubble contours."""

import cv2 as cv
import numpy as np
import pandas as pd
from loguru import logger

from boilercv.data import VIDEO
from boilercv.data.sets import ALL_NAMES, get_dataset
from boilercv.images import scale_bool
from boilercv.images.cv import find_contours
from boilercv.models.params import PARAMS
from boilercv.types import DF, Vid


def main():
    logger.info("start contours")
    for source_name in ALL_NAMES:
        destination = PARAMS.paths.contours / f"{source_name}.h5"
        if destination.exists():
            continue
        video = cv.bitwise_not(scale_bool(get_dataset(source_name)[VIDEO].values))
        df = get_all_contours(video, method=cv.CHAIN_APPROX_SIMPLE)
        df.to_hdf(
            destination,
            "contours",
            complib="zlib",
            complevel=9,
        )
    logger.info("finish contours")


def get_all_contours(video: Vid, method) -> DF:
    """Get all contours."""
    try:
        all_contours = np.vstack(
            [
                np.insert(
                    np.vstack(
                        [
                            np.insert(contour, 0, cont_num, axis=1)
                            for cont_num, contour in enumerate(
                                find_contours(image, method)
                            )
                        ]
                    ),
                    0,
                    image_num,
                    axis=1,
                )
                for image_num, image in enumerate(video)
            ]
        )
    except ValueError:
        logger.exception("no contours found")
        all_contours = np.empty((0, 4))
    return pd.DataFrame(
        all_contours, columns=["frame", "contour", "ypx", "xpx"]
    ).set_index(["frame", "contour"])


if __name__ == "__main__":
    main()



"""Parameters for the data pipeline."""

from typing import Self

from pydantic import Field

from boilercv.models import PARAMS_FILE, MyBaseModel, load_config
from boilercv.models.paths import LocalPaths, Paths


class Params(MyBaseModel):
    """Project parameters."""

    paths: Paths = Field(default_factory=Paths)

    @classmethod
    def get_params(cls: type[Self]) -> Self:
        return load_config(PARAMS_FILE, cls)


PARAMS = Params.get_params()
LOCAL_PATHS = LocalPaths()



"""Paths for this project."""

from pathlib import Path
from typing import Any

from pydantic import DirectoryPath, FilePath, validator
from ruamel.yaml import YAML

from boilercv.models import PARAMS_FILE, MyBaseModel


def init():
    """Synchronize project paths. Run on initial import of this module."""
    yaml = YAML()
    yaml.indent(offset=2)
    paths = Paths()
    params = yaml.load(paths.file_params) or {}
    params["paths"] = repl_path(paths.dict(exclude_none=True))
    yaml.dump(params, paths.file_params)


def repl_path(dirs_dict: dict[str, Path]):
    """Replace Windows path separator with POSIX separator."""
    return {k: str(v).replace("\\", "/") for k, v in dirs_dict.items()}


def get_sorted_paths(path: Path) -> list[Path]:
    """Iterate over a sorted directory."""
    return sorted(path.iterdir())


class LocalPaths(MyBaseModel):
    """Local paths for larger files not stored in the cloud."""

    data: DirectoryPath = Path("~").expanduser() / ".local/boilercv"
    hierarchical_data: DirectoryPath = data / "data"

    large_examples: DirectoryPath = data / "large_examples"
    large_sources: DirectoryPath = data / "large_sources"
    notes: DirectoryPath = data / "notes"
    sheets: DirectoryPath = data / "sheets"
    uncompressed_contours: DirectoryPath = data / "uncompressed_contours"
    uncompressed_filled: DirectoryPath = data / "uncompressed_filled"
    uncompressed_sources: DirectoryPath = data / "uncompressed_sources"

    cines: DirectoryPath = data / "cines"
    large_example_cine: Path = cines / "2022-01-06T16-57-31.cine"

    media: Path = Path("G:/My Drive/Blake/School/Grad/Reports/Content/boilercv")

    # "always" so it'll run even if not in YAML
    # "pre" because dir must exist pre-validation
    @validator(
        "hierarchical_data",
        "large_examples",
        "large_sources",
        "notes",
        "sheets",
        "uncompressed_contours",
        "uncompressed_filled",
        "uncompressed_sources",
        "cines",
        always=True,
        pre=True,
    )
    def validate_output_directories(cls, directory: Path) -> Path:
        """Re-create designated output directories each run, for reproducibility."""
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)
        return directory


class Paths(MyBaseModel):
    """Directories relevant to the project."""

    class Config(MyBaseModel.Config):
        @staticmethod
        def schema_extra(schema: dict[str, Any]):
            for prop in schema.get("properties", {}).values():
                default = prop.get("default")
                if isinstance(default, str):
                    prop["default"] = default.replace("\\", "/")

    # ! PARAMS FILE
    file_params: FilePath = PARAMS_FILE

    # ! REQUIREMENTS
    requirements: FilePath = Path("requirements.txt")
    dev_requirements: DirectoryPath = Path(".tools/requirements")

    # ! PACKAGE
    package: DirectoryPath = Path("src") / "boilercv"
    stages: DirectoryPath = package / "stages"
    models: DirectoryPath = package / "models"
    paths_module: FilePath = models / "paths.py"

    # ! STAGES
    stage_check_cv: FilePath = stages / "check_cv.py"
    stage_contours: FilePath = stages / "contours.py"
    stage_fill: FilePath = stages / "fill.py"
    stage_schema: FilePath = stages / "schema.py"

    # ! PREVIEW STAGES
    update_previews: DirectoryPath = stages / "update_previews"
    stage_binarized_preview: FilePath = update_previews / "binarized.py"
    stage_gray_preview: FilePath = update_previews / "gray.py"
    stage_filled_preview: FilePath = update_previews / "filled.py"

    # ! DATA
    data: DirectoryPath = Path("data")
    contours: DirectoryPath = data / "contours"
    examples: DirectoryPath = data / "examples"
    filled: DirectoryPath = data / "filled"
    rois: DirectoryPath = data / "rois"
    samples: DirectoryPath = data / "samples"
    sources: DirectoryPath = data / "sources"

    # ! PREVIEW DATA
    previews: DirectoryPath = data / "previews"
    binarized_preview: Path = previews / "binarized.nc"
    gray_preview: Path = previews / "gray.nc"
    filled_preview: Path = previews / "filled.nc"

    # ! SCHEMA
    # Can't be "schema", which is a special member of BaseClass
    project_schema: DirectoryPath = data / "schema"

    # "always" so it'll run even if not in YAML
    # "pre" because dir must exist pre-validation
    @validator(
        "contours",
        "examples",
        "filled",
        "rois",
        "samples",
        "sources",
        "previews",
        "project_schema",
        always=True,
        pre=True,
    )
    def validate_output_directories(cls, directory: Path) -> Path:
        """Re-create designated output directories each run, for reproducibility."""
        directory = Path(directory)
        directory.mkdir(parents=True, exist_ok=True)
        return directory


# * -------------------------------------------------------------------------------- * #

init()



"""Fill bubble contours."""

import cv2 as cv
import numpy as np
import pandas as pd
import xarray as xr
from loguru import logger

from boilercv import PREVIEW
from boilercv.captivate.previews import view_images
from boilercv.data import IDX, ROI, VIDEO
from boilercv.data.packing import pack, unpack
from boilercv.data.sets import get_contours_df, get_dataset
from boilercv.examples import EXAMPLE_NUM_FRAMES, EXAMPLE_VIDEO_NAME
from boilercv.images.cv import draw_contours
from boilercv.models.params import PARAMS
from boilercv.types import ArrInt

TRY_EMPTY = False


def main():
    if TRY_EMPTY:
        all_contours = np.empty((0, 4))
        df = pd.DataFrame(
            all_contours, columns=["frame", "contour", "ypx", "xpx"]
        ).set_index(["frame", "contour"])
    else:
        df = get_contours_df(EXAMPLE_VIDEO_NAME)
    ds = xr.zeros_like(
        get_dataset(EXAMPLE_VIDEO_NAME, EXAMPLE_NUM_FRAMES), dtype=np.uint8
    )
    video = ds[VIDEO]
    if not df.empty:
        for frame_num, frame in enumerate(video):
            contours: list[ArrInt] = list(  # type: ignore
                df.loc[IDX[frame_num], :]  # type: ignore
                .groupby("contour")
                .apply(lambda grp: grp.values)  # type: ignore
            )
            video[frame_num, :, :] = draw_contours(
                frame.values, contours, thickness=cv.FILLED
            )
    ds[VIDEO] = pack(video)
    ds = ds.drop_vars(ROI)
    destination = PARAMS.paths.examples / f"{EXAMPLE_VIDEO_NAME}_filled.nc"
    ds.to_netcdf(path=destination)
    if PREVIEW:
        with xr.open_dataset(destination) as ds2:
            view_images([unpack(ds[VIDEO]), unpack(ds2[VIDEO])])


if __name__ == "__main__":
    logger.info("start fill contours")
    main()
    logger.info("finish fill contours")



"""Find blobs in the galaxy."""

from skimage import data
from skimage.color import rgb2gray

from boilercv.captivate.previews import view_images
from boilercv.colors import BLUE, GREEN, RED
from boilercv.examples.blobs import (
    draw_blobs,
    get_blobs_dog,
    get_blobs_doh,
    get_blobs_log,
)
from boilercv.types import ArrInt


def main():
    image = data.hubble_deep_field()[0:500, 0:500]
    image_gray = rgb2gray(image)
    operations = {
        "Laplacian of Gaussian": get_blobs_log,
        "Difference of Gaussian": get_blobs_dog,
        "Determinant of Hessian": get_blobs_doh,
    }
    blobs = {title: func(image_gray) for title, func in operations.items()}
    results: dict[str, ArrInt] = {}
    for (title, blobs_), color in zip(blobs.items(), [RED, GREEN, BLUE], strict=True):
        results[title] = image.copy()
        for blob in blobs_:
            draw_blobs(results[title], blob, color)
    view_images({"input": image} | results)


if __name__ == "__main__":
    main()



"""Update previews for the binarization stage."""

from loguru import logger

from boilercv.data import FRAME, ROI, VIDEO
from boilercv.data.sets import get_dataset
from boilercv.models.params import PARAMS
from boilercv.stages.update_previews import new_videos_to_preview


def main():
    stage = "sources"
    destination = PARAMS.paths.binarized_preview
    with new_videos_to_preview(destination) as videos_to_preview:
        for video_name in videos_to_preview:
            ds = get_dataset(video_name, stage=stage, num_frames=1)
            first_frame = ds[VIDEO].isel({FRAME: 0}).values
            videos_to_preview[video_name] = first_frame & ds[ROI].values


if __name__ == "__main__":
    logger.info("Start update binarized preview")
    main()
    logger.info("Finish update binarized preview")



"""Compress the video data in a NetCDF file.

Compression is slow for grayscale files of ~2GB size, but may work better for binarized
files.
"""

from boilercv import DEBUG
from boilercv.data import VIDEO
from boilercv.examples.large import example_dataset


@example_dataset(  # type: ignore
    destination="compressed", preview=DEBUG, encoding={VIDEO: {"zlib": True}}
)
def main():
    pass


if __name__ == "__main__":
    main()



"""Convert all CINEs to NetCDF."""

from loguru import logger

from boilercv.data.video import prepare_dataset
from boilercv.models.params import LOCAL_PATHS
from boilercv.models.paths import get_sorted_paths


def main():
    logger.info("start convert")
    for source in get_sorted_paths(LOCAL_PATHS.cines):
        destination = LOCAL_PATHS.large_sources / f"{source.stem}.nc"
        if destination.exists():
            continue
        try:
            prepare_dataset(source).to_netcdf(path=destination)
        except Exception:
            logger.exception(source.stem)
            continue
    logger.info("finish convert")


if __name__ == "__main__":
    main()



"""Given a CINE, find ROI using `pyqtgraph` and find contours."""

from matplotlib.pyplot import subplot_mosaic

from boilercv.captivate.previews import edit_roi, load_roi, view_images
from boilercv.examples import EXAMPLE_FRAME_LIST, EXAMPLE_ROI
from boilercv.images import scale_bool
from boilercv.images.cv import (
    apply_mask,
    binarize,
    build_mask_from_polygons,
    draw_contours,
    find_contours,
)
from boilercv.types import ArrInt, Img, ImgBool


def main():
    preview_contours(
        block_size=11,
        thresh_dist_from_mean=2,
        contour_index=-1,
        thickness=2,
    )


def preview_contours(
    block_size: int,
    thresh_dist_from_mean: int,
    contour_index: int,
    thickness: int,
    interact: bool = False,
) -> list[list[ArrInt]] | None:
    input_images = [EXAMPLE_FRAME_LIST[0]] if interact else EXAMPLE_FRAME_LIST
    if interact:
        roi = load_roi(input_images[0], EXAMPLE_ROI)
    else:
        roi = edit_roi(input_images[0], EXAMPLE_ROI)
    all_contours: list[list[ArrInt]] = []
    all_masked: list[Img] = []
    all_thresholded: list[ImgBool] = []
    contoured: list[Img] = []
    to_preview = dict(
        input_images=input_images,
        all_masked=all_masked,
        all_thresholded=all_thresholded,
        contoured=contoured,
    )
    for image in input_images:
        contours, masked, thresholded = _get_contours(
            image, roi, block_size, thresh_dist_from_mean
        )
        all_contours.append(contours)
        all_masked.append(masked)
        all_thresholded.append(thresholded)
        contoured.append(draw_contours(image, contours, contour_index, thickness))
    if interact:
        interact_with_images(*[image[0] for image in to_preview.values()])
    else:
        view_images(to_preview)
        return all_contours


def _get_contours(
    input_image: ArrInt,
    roi: ArrInt,
    block_size: int,
    thresh_dist_from_mean: int,
) -> tuple[list[ArrInt], ArrInt, ImgBool]:
    masked = apply_mask(input_image, build_mask_from_polygons(input_image, [roi]))
    thresholded = binarize(masked, block_size, thresh_dist_from_mean)
    return find_contours(scale_bool(~thresholded)), masked, thresholded


def interact_with_images(_input_image, _masked, thresholded, contoured):
    _plt, ax = subplot_mosaic(
        [
            [
                # "input",
                # "masked",
                "thresholded",
                "contoured",
            ]
        ]
    )
    for ax_ in ax.values():  # type: ignore
        ax_.axis("off")
    # ax["input"].imshow(input_image, cmap="gray")  # type: ignore
    # ax["masked"].imshow(masked, cmap="gray")  # type: ignore
    ax["thresholded"].imshow(thresholded, cmap="gray")  # type: ignore
    ax["contoured"].imshow(contoured, cmap="gray")  # type: ignore


if __name__ == "__main__":
    main()



"""Find bubbles as blobs."""

import cv2 as cv

from boilercv.captivate.previews import edit_roi, view_images
from boilercv.colors import RED
from boilercv.examples import EXAMPLE_FRAME_LIST, EXAMPLE_ROI
from boilercv.examples.blobs import draw_blobs, get_blobs_doh
from boilercv.images.cv import apply_mask, build_mask_from_polygons, convert_image
from boilercv.types import ArrInt

_NUM_FRAMES = 10
SHORTER_FRAME_LIST = EXAMPLE_FRAME_LIST[:_NUM_FRAMES]


def main():
    roi = edit_roi(SHORTER_FRAME_LIST[0], EXAMPLE_ROI)
    # results_log: list[ArrInt] = []
    # results_dog: list[ArrInt] = []
    results_doh: list[ArrInt] = []
    all_results = [
        # results_log,
        # results_dog,
        results_doh,
    ]
    for input_image in SHORTER_FRAME_LIST:
        image = apply_mask(input_image, build_mask_from_polygons(input_image, [roi]))
        all_blobs = [
            # get_blobs_log(image),
            # get_blobs_dog(image),
            get_blobs_doh(image),
        ]
        sequence = zip(all_blobs, all_results, strict=True)
        for blobs, results in sequence:
            result = convert_image(input_image, cv.COLOR_GRAY2RGB)
            for blob in blobs:
                draw_blobs(result, blob, RED)
            results.append(result)
    view_images([SHORTER_FRAME_LIST, *all_results])


if __name__ == "__main__":
    main()



"""Update previews for various stages."""

from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

import xarray as xr

from boilercv.captivate.previews import pad_images
from boilercv.data import VIDEO, VIDEO_NAME, XPX, YPX, assign_ds
from boilercv.data.models import Dimension
from boilercv.data.sets import ALL_NAMES
from boilercv.types import DS


@contextmanager
def new_videos_to_preview(
    destination: Path, rebuild: bool = False
) -> Iterator[dict[str, Any]]:
    """Get an empty mapping of new videos to preview and write to disk."""

    # Yield a mapping of new video names to previews, to be populated by the user
    if rebuild:
        new_video_names = ALL_NAMES
    else:
        existing_names: list[str] = []
        if destination.exists():
            with xr.open_dataset(destination) as existing_ds:
                existing_names.extend(list(existing_ds[VIDEO_NAME].values))
        new_video_names = [name for name in ALL_NAMES if name not in existing_names]
    videos_to_preview = dict.fromkeys(new_video_names)

    yield videos_to_preview

    # Keep only valid received previews and build a dataset of new previews
    if received_previews := {
        video_name: preview
        for video_name, preview in videos_to_preview.items()
        if preview is not None and video_name in new_video_names
    }:
        received_video_names = list(received_previews.keys())
        received_previews = list(received_previews.values())
        new_ds = get_preview_ds(received_video_names, received_previews)

        if destination.exists():
            with xr.open_dataset(destination) as existing_ds:
                if new_ds[VIDEO].shape == existing_ds[VIDEO].shape:
                    # Combine datasets if they're the same shape
                    new_ds = xr.combine_by_coords([existing_ds, new_ds])
                else:
                    # Otherwise, destructure, pad, and rebuild them together
                    existing_names = list(existing_ds[VIDEO_NAME].values)
                    new_ds = get_preview_ds(
                        existing_names + received_video_names,
                        received_previews + list(existing_ds.video.values),
                    )

        new_ds.to_netcdf(path=destination, encoding={VIDEO: {"zlib": True}})


def get_preview_ds(preview_names: list[str], previews: list[Any]) -> DS:
    """Get a dataset of preview images, padding sizes as necessary."""
    return assign_ds(
        name=VIDEO,
        long_name="Video preview",
        units="Pixel state",
        dims=(
            Dimension(
                dim=VIDEO_NAME,
                long_name="Video name",
                coords=preview_names,
            ),
            Dimension(
                dim=YPX,
                long_name="Height",
                units="px",
            ),
            Dimension(
                dim=XPX,
                long_name="Width",
                units="px",
            ),
        ),
        data=pad_images(previews),
    )



"""Datasets."""

from collections.abc import Iterator
from pathlib import Path
from typing import Literal, TypeAlias

import pandas as pd
import xarray as xr

from boilercv.data import HEADER, ROI, VIDEO
from boilercv.data.packing import unpack
from boilercv.models.params import LOCAL_PATHS, PARAMS
from boilercv.models.paths import get_sorted_paths
from boilercv.types import DF, DS

ALL_FRAMES = slice(None)
ALL_SOURCES = get_sorted_paths(PARAMS.paths.sources)
ALL_NAMES = [source.stem for source in ALL_SOURCES]

Stage: TypeAlias = Literal["large_sources", "sources", "filled"]
STAGE_DEFAULT = "sources"


def get_all_datasets(
    num_frames: int = 0, frame: slice = ALL_FRAMES, stage: Stage = STAGE_DEFAULT
) -> Iterator[tuple[DS, str]]:
    """Yield datasets in order."""
    for name in ALL_NAMES:
        yield get_dataset(name, num_frames, frame, stage), name


def inspect_dataset(name: str, stage: Stage = STAGE_DEFAULT) -> DS:
    """Inspect a video dataset."""
    cmp_source, unc_source = get_stage(name, stage)
    source = unc_source if unc_source.exists() else cmp_source
    if stage == "large_sources":
        return xr.open_dataset(source) if source.exists() else xr.Dataset()
    with xr.open_dataset(source) as ds:
        return ds


def get_dataset(
    name: str,
    num_frames: int = 0,
    frame: slice = ALL_FRAMES,
    stage: Stage = STAGE_DEFAULT,
) -> DS:
    """Load a video dataset."""
    # Can't use `xr.open_mfdataset` because it requires dask
    # Unpacking is incompatible with dask
    frame = slice_frames(num_frames, frame)
    cmp_source, unc_source = get_stage(name, stage)
    source = unc_source if unc_source.exists() else cmp_source
    if stage == "large_sources":
        ds = xr.open_dataset(source)
        return (
            xr.Dataset({VIDEO: ds[VIDEO].sel(frame=frame), HEADER: ds[HEADER]})
            if source.exists()
            else xr.Dataset()
        )
    roi = PARAMS.paths.rois / f"{name}.nc"
    with xr.open_dataset(source) as ds, xr.open_dataset(roi) as roi_ds:
        if not unc_source.exists():
            xr.Dataset({VIDEO: ds[VIDEO], HEADER: ds[HEADER]}).to_netcdf(
                path=unc_source, encoding={VIDEO: {"zlib": False}}
            )
        return xr.Dataset(
            {
                VIDEO: unpack(ds[VIDEO].sel(frame=frame)),
                ROI: roi_ds[ROI],
                HEADER: ds[HEADER],
            }
        )


def get_stage(name: str, stage: Stage = STAGE_DEFAULT) -> tuple[Path, Path]:
    """Get the paths associated with a particular video name and pipeline stage."""
    if stage == "sources":
        unc_source = LOCAL_PATHS.uncompressed_sources / f"{name}.nc"
        return PARAMS.paths.sources / f"{name}.nc", unc_source
    elif stage == "large_sources":
        source = unc_source = LOCAL_PATHS.large_sources / f"{name}.nc"
        return source, unc_source
    elif stage == "filled":
        unc_source = LOCAL_PATHS.uncompressed_filled / f"{name}.nc"
        return PARAMS.paths.filled / f"{name}.nc", unc_source
    else:
        raise ValueError(f"Unknown stage: {stage}")


def get_contours_df(name: str) -> DF:
    """Load contours from a dataset."""
    unc_cont = LOCAL_PATHS.uncompressed_contours / f"{name}.h5"
    contour = unc_cont if unc_cont.exists() else PARAMS.paths.contours / f"{name}.h5"
    contour_df: DF = pd.read_hdf(contour)  # type: ignore
    if not unc_cont.exists():
        contour_df.to_hdf(unc_cont, key="contours", complevel=None, complib=None)
    return contour_df


def slice_frames(num_frames: int = 0, frame: slice = ALL_FRAMES) -> slice:
    """Returns a slice suitable for getting frames from datasets."""
    if num_frames:
        if frame == ALL_FRAMES:
            frame = slice(None, num_frames - 1)
        else:
            raise ValueError("Don't specify both `num_frames` and `frame`.")
    return frame



"""Image and video previews."""

import inspect
from collections.abc import Callable, Mapping, MutableSequence, Sequence
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Literal, TypeAlias

import numpy as np
import pandas as pd
import pyqtgraph as pg
import yaml
from PySide6.QtCore import QEvent, Qt, Signal
from PySide6.QtGui import QKeyEvent
from PySide6.QtWidgets import QGridLayout, QHBoxLayout, QLayout, QPushButton

from boilercv import FRAMERATE_CONT
from boilercv.images import scale_bool
from boilercv.types import DA, ArrInt, Img


def init():
    """Initialize `boilercv.gui`."""
    pg.setConfigOption("imageAxisOrder", "row-major")


Viewable: TypeAlias = Any  # The true type is a complex union of lots of array types
NamedViewable: TypeAlias = Mapping[str | int, Viewable]
MultipleViewable: TypeAlias = Sequence[Viewable]
MutableViewable: TypeAlias = MutableSequence[Viewable]
AllViewable: TypeAlias = Viewable | NamedViewable | MultipleViewable

# * -------------------------------------------------------------------------------- * #
# * IMAGE VIEWER

WINDOW_SIZE = (800, 600)
WINDOW_NAME = "Image viewer"
SMALLER_GRIDS = {
    1: (1, 1),
    2: (1, 2),
    3: (1, 3),
    4: (2, 2),
    5: (2, 3),
    6: (2, 3),
    7: (2, 4),
    8: (2, 4),
    9: (2, 5),
    10: (2, 5),
    11: (3, 4),
    12: (3, 4),
}
"""Rectangular grid sizes for up to twelve views. Use square grids beyond that."""


def view_images(images: AllViewable, name: str = "", framerate: int = FRAMERATE_CONT):
    """Compare multiple images or videos."""
    with image_viewer(images, name, framerate):
        pass


@contextmanager
def image_viewer(images: AllViewable, name: str = "", framerate: int = FRAMERATE_CONT):  # type: ignore  # noqa: C901
    """View and interact with images and video."""

    images: NamedViewable = coerce_images(images)
    num_views = len(images)
    shape = SMALLER_GRIDS.get(num_views, get_square_grid(num_views))
    height, width = shape
    coords = get_grid_coordinates(shape)
    image_views, app, window, layout, button_layout = make_window(name)

    def main():
        """Handle required setup and teardown."""
        add_image_views()
        add_actions()
        image_view_mapping = set_images(images, image_views)
        try:
            yield image_view_mapping, app, window, layout, button_layout
        finally:
            window.show()
            app.exec()

    def add_image_views():
        """Add image views in a grid."""
        for column in range(width):
            layout.setColumnStretch(column, 1)
        for coord in coords:
            image_view = get_image_view(framerate)
            layout.addWidget(image_view, *coord)
            image_views.append(image_view)

    def add_actions():
        """Add buttons and keypress events to the window."""
        window.key_signal.connect(keyPressEvent)
        layout.addLayout(button_layout, height, 0, 1, width)
        if num_views > 1:
            add_button(button_layout, "Toggle Play All", trigger_space).setFocus()
        add_button(button_layout, "Continue", app.quit)

    def trigger_space():
        """Trigger a spacebar keypress event. For propagating logic to image views."""
        keyPressEvent(
            QKeyEvent(
                QEvent.Type.KeyPress, Qt.Key.Key_Space, Qt.KeyboardModifier.NoModifier
            )
        )

    def keyPressEvent(ev: QKeyEvent):  # noqa: N802
        """Handle quit events and propagate keypresses to image views."""
        if ev.key() in (Qt.Key.Key_Escape, Qt.Key.Key_Q, Qt.Key.Key_Enter):
            app.quit()
            ev.accept()
        for image_view in image_views:
            image_view.keyPressEvent(ev)

    yield from main()


# * -------------------------------------------------------------------------------- * #


def make_window(name: str = ""):
    """Start a PyQtGraph app and prepare an app window with a default layout."""
    # Isolate pg.ImageView in a layout cell. It is complicated to directly modify the UI
    # of pg.ImageView. Can't use the convenient pg.LayoutWidget because
    # GraphicsLayoutWidget cannot contain it, and GraphicsLayoutWidget is convenient on
    # its own.
    app = pg.mkQApp()
    image_views: list[pg.ImageView] = []
    layout = QGridLayout()
    button_layout = QHBoxLayout()
    window = GraphicsLayoutWidgetWithKeySignal(size=WINDOW_SIZE)
    window.setLayout(layout)
    window.setWindowTitle(f"{WINDOW_NAME}: {name or get_calling_scope_name()}")
    return image_views, app, window, layout, button_layout


def get_grid_coordinates(shape: tuple[int, int]) -> list[tuple[int, int]]:
    """Get the coordinates of a grid."""
    (y, x) = np.ones(shape).nonzero()
    return list(zip(y, x, strict=True))


def get_square_grid(num_views: int) -> tuple[int, int]:
    """Get the dimensions of a square grid for the given number of views."""
    minimum_side_length = np.ceil(np.sqrt(num_views))
    side_length = int(minimum_side_length)
    (height, width) = (side_length,) * 2
    return height, width


def coerce_images(images: AllViewable) -> NamedViewable:
    """Coerce images to a mapping of title to image."""
    if isinstance(images, Mapping):
        images_ = images
    elif isinstance(images, np.ndarray | DA):
        images_ = [images]
    elif isinstance(images, Sequence):
        # If given a sequence that could be a video or a set of images/videos to
        # compare, assume it is a video if it is too long to be a set of comparisons.
        largest_grid = 16
        if len(images) > largest_grid:
            images_ = [np.array(pad_images(images))]
        else:
            images_ = images
    else:
        raise TypeError(f"Unsupported type for images: {type(images)}")

    return (
        {title: np.array(pad_images(viewable)) for title, viewable in images_.items()}
        if isinstance(images_, Mapping)
        else {i: np.array(pad_images(viewable)) for i, viewable in enumerate(images_)}
    )


def pad_images(images: MultipleViewable) -> MutableViewable:  # type: ignore
    """Pad images to a common size and pack into an array."""
    flat_image = isinstance(images, np.ndarray | DA) and (
        # One-channel
        images.ndim == 2  # type: ignore
        # Up to four-channel
        or images.ndim == 3  # type: ignore
        and images.shape[-1] <= 4  # type: ignore
    )
    images: MutableViewable = [images] if flat_image else list(images)
    shapes = pd.DataFrame(
        columns=["height", "width"], data=list({image.shape[:2] for image in images})
    ).set_index(["height", "width"], drop=False)
    if len(shapes) == 1:
        return images

    pads = (
        (shapes[["height", "width"]].max() - shapes[["height", "width"]]) // 2
    ).set_axis(axis="columns", labels=["hpad", "wpad"])
    for i, image in enumerate(images):
        pad: tuple[int, int] = pads.loc[image.shape[:2], :]  # type: ignore
        hpad, wpad = pad
        zero_pad_for_additional_dims = ((0, 0),) * (image.ndim - 2)
        pad_width = ((hpad, hpad), (wpad, wpad), *zero_pad_for_additional_dims)
        images[i] = np.pad(image, pad_width)
    return images


def set_images(
    images: NamedViewable, image_views: list[pg.ImageView]
) -> dict[str | int, pg.ImageView]:
    """Set images into the image views."""
    for (title, viewable), image_view in zip(images.items(), image_views, strict=False):
        if np.issubdtype(viewable.dtype, bool):
            viewable = scale_bool(viewable)
        image_view.setImage(viewable.squeeze())
        if isinstance(title, str):
            image_view.addItem(pg.TextItem(title, fill=pg.mkBrush("black")))
    return dict(zip(images.keys(), image_views, strict=False))


class GraphicsLayoutWidgetWithKeySignal(pg.GraphicsLayoutWidget):
    """Emit key signals on `key_signal`."""

    key_signal = Signal(QKeyEvent)

    def keyPressEvent(self, ev: QKeyEvent):  # noqa: N802
        super().keyPressEvent(ev)
        self.key_signal.emit(ev)


def add_button(layout: QLayout, label: str, func: Callable[..., Any]) -> QPushButton:
    """Add a named button to a layout and connect it to a callback."""
    button = QPushButton(label)
    button.clicked.connect(func)  # type: ignore
    layout.addWidget(button)
    return button


def get_image_view(framerate: int = FRAMERATE_CONT) -> pg.ImageView:
    """Get an image view suitable for previewing images."""
    image_view = pg.ImageView()
    image_view.playRate = framerate
    image_view.ui.histogram.hide()
    image_view.ui.roiBtn.hide()
    image_view.ui.menuBtn.hide()
    return image_view


def get_calling_scope_name():
    """Get the name of the calling scope."""
    current_frame = inspect.currentframe()
    scope_name = current_frame.f_back.f_code.co_name  # type: ignore
    while scope_name in {
        "image_viewer",
        "view_images",
        "preview_images",
        "make_window",
        "__enter__",
        "eval_in_context",
        "evaluate_expression",
        "_run_with_interrupt_thread",
        "_run_with_unblock_threads",
        "new_func",
        "internal_evaluate_expression_json",
        "do_it",
        "process_internal_commands",
    }:
        current_frame = current_frame.f_back  # type: ignore
        scope_name = current_frame.f_back.f_code.co_name  # type: ignore
    return scope_name


# * -------------------------------------------------------------------------------- * #


def save_roi(roi_vertices: ArrInt, roi_path: Path):
    """Save an ROI represented by an ordered array of vertices."""
    roi_path.write_text(
        encoding="utf-8", data=yaml.safe_dump(roi_vertices.tolist(), indent=2)
    )


def edit_roi(
    image: ArrInt, roi_path: Path, roi_type: Literal["poly", "line"] = "poly"
) -> ArrInt:
    """Edit the region of interest for an image."""

    with image_viewer(image) as (image_views, _app, window, _layout, button_layout):
        common_roi_args = dict(
            pen=pg.mkPen("red"),
            hoverPen=pg.mkPen("magenta"),
            handlePen=pg.mkPen("blue"),
            handleHoverPen=pg.mkPen("magenta"),
            positions=np.fliplr(load_roi(image, roi_path, roi_type)),
        )
        roi = (
            pg.PolyLineROI(
                **common_roi_args,
                closed=True,
            )
            if roi_type == "poly"
            else pg.LineSegmentROI(**common_roi_args)
        )

        def main():
            """Allow ROI interaction."""
            window.key_signal.connect(keyPressEvent)
            button = QPushButton("Save ROI")
            button.clicked.connect(save_roi_)  # type: ignore
            button_layout.addWidget(button)
            image_views[0].addItem(roi)

        def keyPressEvent(ev: QKeyEvent):  # noqa: N802
            """Save ROI or quit on key presses."""
            if ev.key() == Qt.Key.Key_S:
                save_roi_()
                ev.accept()

        def save_roi_():
            """Save the ROI."""
            vertices = get_roi_vertices()
            save_roi(vertices, roi_path)

        def get_roi_vertices() -> ArrInt:
            """Get the vertices of the ROI."""
            return np.fliplr(np.array(roi.saveState()["points"], dtype=int))

        main()

    return get_roi_vertices()


def load_roi(
    img: Img,
    roi_path: Path,
    roi_type: Literal["poly", "line"] = "poly",
) -> ArrInt:
    """Load the region of interest for an image."""
    (width, height) = img.shape[-2:]
    if roi_path.exists():
        vertices: list[tuple[int, int]] = yaml.safe_load(
            roi_path.read_text(encoding="utf-8")
        )
    else:
        vertices = (
            [(0, 0), (0, width), (height, width), (height, 0)]
            if roi_type == "poly"
            else [(0, 0), (height, width)]
        )
    return np.array(vertices, dtype=int)


# * -------------------------------------------------------------------------------- * #

init()



"""Given an MP4, find ROI using `opencv` and find contours."""


import cv2 as cv
import numpy as np
from scipy.spatial import ConvexHull

from boilercv.colors import BLUE_CV
from boilercv.examples.cv import capture_images
from boilercv.images import scale_bool
from boilercv.images.cv import (
    apply_mask,
    binarize,
    build_mask_from_polygons,
    convert_image,
    draw_contours,
    find_contours,
)
from boilercv.models.params import PARAMS
from boilercv.types import ArrInt

WINDOW_NAME = "image"
ESC_KEY = ord("\x1b")


def main():
    images = (
        image[:, :, 0]
        for image in capture_images(
            PARAMS.paths.examples / "2022-04-08T16-12-42_short.mp4"
        )
    )
    roi = np.fliplr(get_roi2(next(images)))
    for image in images:
        masked = apply_mask(image, build_mask_from_polygons(image, [roi]))
        thresholded = binarize(masked)
        contours = find_contours(scale_bool(~thresholded))
        image_with_contours = draw_contours(image, contours)
        cv.imshow(WINDOW_NAME, image_with_contours)
        if cv.waitKey(100) == ESC_KEY:
            break
    cv.destroyAllWindows()


def get_roi2(image: ArrInt) -> ArrInt:
    """Get the region of interest of an image.

    See: https://docs.opencv.org/4.6.0/db/d5b/tutorial_py_mouse_handling.html
    """

    clicks: list[tuple[int, int]] = []

    image = convert_image(image, cv.COLOR_GRAY2RGB)
    (width, height) = image.shape[:-1]
    click = 0
    default_clicks = [(0, 0), (0, width), (height, width), (height, 0)]
    hull = ConvexHull(default_clicks)
    hull_minimum_vertices = 3
    composite_image = image

    def main() -> ArrInt:
        nonlocal clicks
        cv.imshow(WINDOW_NAME, image)
        cv.setMouseCallback(WINDOW_NAME, handle_mouse_events)
        while True and cv.waitKey(100) != ESC_KEY:
            pass
        if len(clicks) < hull_minimum_vertices:
            clicks = default_clicks
        else:
            hull.close()
        return np.array(clicks)[hull.vertices]

    def handle_mouse_events(event: int, x: int, y: int, *_):
        """Handle all mouse events. Form a convex hull from left clicks."""
        nonlocal image, click, hull, composite_image
        if event == cv.EVENT_LBUTTONDOWN:
            click += 1
            clicks.append((x, y))
            image = cv.drawMarker(image, clicks[-1], BLUE_CV)
            if click == hull_minimum_vertices:
                hull = ConvexHull(clicks, incremental=True)
                composite_image = draw_hull(hull, image)
            elif click > hull_minimum_vertices:
                hull.add_points([clicks[-1]])
                composite_image = draw_hull(hull, image)
            cv.imshow(WINDOW_NAME, composite_image)

    def draw_hull(hull: ConvexHull, image: ArrInt) -> ArrInt:
        """Draw a convex hull on the image."""
        image = image.copy()
        clicks = hull.points.astype(int)
        for simplex in hull.simplices:
            image = cv.line(image, clicks[simplex[0]], clicks[simplex[1]], BLUE_CV)
        return image

    return main()


if __name__ == "__main__":
    main()



from dataclasses import dataclass
from math import sqrt

from skimage.draw import disk, set_color
from skimage.feature import blob_dog, blob_doh, blob_log


def get_blobs_dog(image):
    blobs = blob_dog(
        image,
        # max_sigma=30,
        # threshold=0.1,
    )
    blobs[:, 2] = blobs[:, 2] * sqrt(2)
    return (Blob(*blob) for blob in blobs)


def get_blobs_log(image):
    blobs = blob_log(
        image,
        max_sigma=30,
        num_sigma=10,
        threshold=0.1,
    )
    blobs[:, 2] = blobs[:, 2] * sqrt(2)
    return (Blob(*blob) for blob in blobs)


def get_blobs_doh(image):
    blobs = blob_doh(
        image,
        # min_sigma=1,
        # max_sigma=30,
        # threshold=0.01,
        # overlap=0.30,
    )
    return (Blob(*blob) for blob in blobs)


@dataclass
class Blob:
    y: int
    x: int
    r: int


def draw_blobs(result, blob, color):
    rr, cc = disk(
        (int(blob.y), int(blob.x)),
        int(blob.r),
    )
    set_color(result, (rr, cc), color)



"""Compress the packed video data in a NetCDF file.

This is much quicker, and yields another 8x compresssion ratio on top of the packing.
"""

from boilercv import DEBUG
from boilercv.data import VIDEO
from boilercv.examples.large import example_dataset


@example_dataset(  # type: ignore
    source="packed",
    destination="packed_compressed",
    preview=DEBUG,
    encoding={VIDEO: {"zlib": True}},
)
def main():
    pass


if __name__ == "__main__":
    main()



"""Examples, experiments, and demonstrations."""

import xarray as xr

from boilercv.data import VIDEO
from boilercv.models.params import PARAMS
from boilercv.types import DA

EXAMPLE_NUM_FRAMES = 1000
EXAMPLE_VIDEO_NAME = "2022-11-30T13-41-00"
EXAMPLE_CONTOURS = PARAMS.paths.examples / f"{EXAMPLE_VIDEO_NAME}.h5"
# TODO: Source the ROI from the dataset.
EXAMPLE_ROI = PARAMS.paths.examples / f"{EXAMPLE_VIDEO_NAME}_roi.yaml"


def get_images() -> DA:
    with xr.open_dataset(PARAMS.paths.examples / f"{EXAMPLE_VIDEO_NAME}.nc") as ds:
        return ds[VIDEO].sel(frame=slice(None, EXAMPLE_NUM_FRAMES))


EXAMPLE_VIDEO = get_images()
EXAMPLE_FRAME_LIST = list(EXAMPLE_VIDEO.values)



"""Compare the original binarized dataset with the round-tripped dataset."""

from boilercv import DEBUG
from boilercv.examples.large import example_dataset as ex_ds

KW = dict(preview=DEBUG, save=False)


def main():
    with ex_ds("binarized", **KW) as original, ex_ds("unpacked", **KW) as unpacked:
        assert unpacked.identical(original)


if __name__ == "__main__":
    main()



"""Convert CINE files to the NetCDF file format."""

from boilercv.data.video import prepare_dataset
from boilercv.examples import EXAMPLE_NUM_FRAMES
from boilercv.models.params import LOCAL_PATHS


def main():
    destination = (
        LOCAL_PATHS.large_examples / f"{LOCAL_PATHS.large_example_cine.stem}.nc"
    )
    ds = prepare_dataset(LOCAL_PATHS.large_example_cine, num_frames=EXAMPLE_NUM_FRAMES)
    ds.to_netcdf(path=destination)


if __name__ == "__main__":
    main()



"""Image and video capturing."""

from pathlib import Path

import imageio
import numpy as np
from loguru import logger

from boilercv import FFMPEG_LOG_LEVEL, FRAMERATE_CONT
from boilercv.images import scale_bool
from boilercv.types import DA, Img, ImgBool, Vid, VidBool


def write_video(
    path: Path,
    video: Vid | VidBool | DA,
    framerate: int = FRAMERATE_CONT,
    preview_frame: bool = False,
):
    """Write a video to disk with the default filetype and timestamp.

    Args:
        path: Path to the video file (suffix coerced to '.mp4').
        video: Data structure to write as a video.
        framerate: Frames per second. Default: Package default framerate.
        preview_frame: Write the first frame to disk as an image. Default: False.
    """
    video = coerce_input(video)
    if path.suffix and path.suffix != ".mp4":
        logger.warning(f"Changing extesion of {path}  to '.mp4'.")
    path = path.with_suffix(".mp4")
    writer = imageio.get_writer(
        uri=path, fps=framerate, macro_block_size=8, ffmpeg_log_level=FFMPEG_LOG_LEVEL
    )
    for image in video:
        writer.append_data(image)
    writer.close()
    if preview_frame:
        write_image(path.with_suffix(".png"), video[0])


def write_image(path: Path, image: Img | ImgBool | DA):
    """Write an image to disk with default filetype.

    Args:
        path: Path to the image file (suffix coerced to '.png').
        image: Data structure to write as an image.
    """
    image = coerce_input(image)
    if path.suffix and path.suffix != ".png":
        logger.warning(f"Changing extesion of {path} to '.png'.")
    path = path.with_suffix(".png")
    imageio.imwrite(path, image)


def coerce_input(imgs: Img | ImgBool | DA) -> Img:
    """Coerce input image or video to the appropriate type.

    Args:
        imgs: Image or video to coerce.
    """
    if np.issubdtype(imgs.dtype, np.integer):
        viewable: Img = imgs.values if isinstance(imgs, DA) else imgs  # type: ignore
    elif np.issubdtype(imgs.dtype, bool):
        viewable: Img = (
            scale_bool(imgs.values) if isinstance(imgs, DA) else scale_bool(imgs)  # type: ignore
        )
    else:
        raise TypeError(f"Cannot coerce {type(imgs)} to a viewable type.")
    return viewable



"""Remove the "results" prefix from files."""

from pathlib import Path


def main():
    folder = Path("C:/Users/Blake/Desktop/Video")
    videos = sorted(folder.iterdir())
    for video in videos:
        new_name = video.name.removeprefix("results_")
        video.rename(folder / new_name)


if __name__ == "__main__":
    main()



"""Store binarized images in a NetCDF file."""

from boilercv.data import VIDEO, apply_to_img_da
from boilercv.examples.large import example_dataset
from boilercv.images.cv import binarize


def main():
    with example_dataset(destination="binarized") as ds:
        ds[VIDEO] = apply_to_img_da(binarize, ds[VIDEO], vectorize=True)


if __name__ == "__main__":
    main()
